/* $Id: libmpi.h,v 1.6 2006/08/02 23:03:56 dalcinl Exp $ */

#ifndef Py_LIBMPI_H
#define Py_LIBMPI_H

/* Below, some stuff from the top of MPICH2 mpicxx.h ...

   There is a name conflict between stdio.h and the MPI C++ binding 
   with respect to the names SEEK_SET, SEEK_CUR, and SEEK_END.  MPI
   wants these in the MPI namespace, but stdio.h will #define these
   to integer values.  #undef'ing these can cause obscure problems
   with other include files (such as iostream), so we instead use
   #error to indicate a fatal error.  Users can either #undef 
   the names before including mpi.h or include mpi.h *before* stdio.h
   or iostream.

   The MPI STANDARD HAS TO BE CHANGED to prevent this nonsense. 
*/

#include <Python.h>
#include <mpi.h>

#ifdef __cplusplus
extern "C" {
#endif

#if !defined(libmpi_MODULE)

#define libmpi_FatalApiError \
(Py_FatalError("Call to API function without first calling " \
	       "import_libmpi(). [file: " __FILE__ "]"), NULL)

#endif /* !libmpi_MODULE */

<MPI_OBJECT>

#if !defined(libmpi_MODULE)

static int import_libmpi(void)
{ 
  PyObject* m = PyImport_ImportModule("mpi4py.libmpi");
  if (m == NULL) return -1;
# define LIBMPI_GET_TYPE_AND_API(m,Type) {               \
  PyObject *tp, *api;                                    \
  tp  = PyObject_GetAttrString(m,#Type"Type");           \
  if (tp == NULL) goto fail;                             \
  PyMPI##Type##_TYPE = (PyTypeObject*) tp;               \
  Py_DECREF(tp);                                         \
  api = PyObject_GetAttrString(m,"PyMPI"#Type"_API");    \
  if (api == NULL) goto fail;                            \
  PyMPI##Type##_API = (void**) PyCObject_AsVoidPtr(api); \
  Py_DECREF(api);                                        \
  if (PyMPI##Type##_API == NULL) goto fail;              \
  }
  LIBMPI_GET_TYPE_AND_API(m, Op         );
  LIBMPI_GET_TYPE_AND_API(m, Comm       );
  LIBMPI_GET_TYPE_AND_API(m, Group      );
  LIBMPI_GET_TYPE_AND_API(m, Status     );
  LIBMPI_GET_TYPE_AND_API(m, Request    );
  LIBMPI_GET_TYPE_AND_API(m, Datatype   );
  LIBMPI_GET_TYPE_AND_API(m, Errhandler );
# undef LIBMPI_GET_TYPE_AND_API
  Py_DECREF(m);
  return 0;
 fail:
  Py_DECREF(m);
  PyErr_Clear();
  PyErr_SetString(PyExc_ImportError,
		  "Can't get MPI types or API from module 'libmpi'");
  return -1;
}

#endif /* !libmpi_MODULE */


#ifdef __cplusplus
}
#endif

#endif /* !Py_LIBMPI_H */

/* 
 * Local Variables:
 * mode: C
 * End:
 */
