/* $Id: libmpi.c,v 1.7 2006/08/08 21:04:29 dalcinl Exp $ */

#define MPICH_IGNORE_CXX_SEEK

#define  libmpi_MODULE
#include "libmpi.h"

/* PyMPIComm_Type implementation */

static PyTypeObject _Py_MPI_Comm_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,				/*ob_size*/
  "_Py_MPI_Comm",		/*tp_name*/
  sizeof(_Py_MPI_Comm),	/*tp_basicsize*/
  0,				/*tp_itemsize*/
  (destructor)PyObject_Del,	/*tp_dealloc*/
  0,
};


#define MPI_COMM_DEFAULT MPI_COMM_NULL

#define comm_alloc PyType_GenericAlloc

static void
comm_dealloc(PyMPICommObject *self)
{
  Py_XDECREF(self->comm);
  self->ob_type->tp_free(self);
}

static PyObject *
comm_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
  PyObject *ob = NULL;
  static char *kwlist[] = {"comm", 0};
  
  if (!PyArg_ParseTupleAndKeywords(args, kwds, 
				   "|O:Comm", kwlist,
				   &ob)) return NULL;
  
  if (ob == NULL || ob == Py_None) {
    PyMPICommObject *self;
    self = (PyMPICommObject*) type->tp_alloc(type, 0);
    if (self != NULL) {
      self->comm = (PyObject*) PyObject_New(_Py_MPI_Comm,
					     &_Py_MPI_Comm_Type);
      if (self->comm == NULL) {
	Py_DECREF(self);
	return NULL;
      }
      PyMPIComm_AS_COMM(self) = MPI_COMM_DEFAULT;
    }
    return (PyObject*)self;
  }
  
  if (ob->ob_type == type) {
    Py_INCREF(ob);
    return ob;
  }

  if (PyMPIComm_Check(ob)) {
    PyMPICommObject *self;
    self = (PyMPICommObject*) type->tp_alloc(type, 0);
    if (self != NULL) {
      self->comm = ((PyMPICommObject*)ob)->comm;
      if (self->comm == NULL) {
	Py_DECREF(self);
	PyErr_BadInternalCall();
	return NULL;
      }
      Py_INCREF(self->comm);
    }
    return (PyObject*) self;
  }
  
  PyErr_SetString(PyExc_TypeError,
		  "Comm() argument must be a 'MPI_Comm' object");
  return NULL;
}

#define comm_init (initproc)0

static PyObject *
comm_richcompare(PyMPICommObject * o1, PyMPICommObject * o2, int op)
{
  PyObject *res;

  /* Make sure both arguments are the same type */
  if (!(PyMPIComm_Check((PyObject*)o1) && 
	PyMPIComm_Check((PyObject*)o2))) {
    Py_INCREF(Py_NotImplemented);
    return Py_NotImplemented;
  }
  if (op != Py_EQ && op != Py_NE) {
    PyErr_SetString(PyExc_TypeError,
		    "cannot compare 'MPI_Comm' objects using <, <=, >, >=");
    return NULL;
  }

  if ((op == Py_EQ) == 
      (PyMPIComm_AS_COMM(o1) == PyMPIComm_AS_COMM(o2)))
    res = Py_True;
  else
    res = Py_False;
  
  Py_INCREF(res);
  return res;
}


static int
comm_nonzero(PyMPICommObject* ob) {
  return PyMPIComm_AS_COMM(ob) != MPI_COMM_NULL;
}
static PyNumberMethods comm_number_methods = {
	0,			/*nb_add*/
	0,			/*nb_subtract*/
	0,			/*nb_multiply*/
	0, 	 	 	/*nb_divide*/
	0,			/*nb_remainder*/
	0,			/*nb_divmod*/
	0,			/*nb_power*/
	0,			/*nb_negative*/
	0,			/*nb_positive*/
	0,			/*nb_absolute*/
	(inquiry)comm_nonzero,	/*nb_nonzero*/
	0,			/*nb_invert*/
	0,			/*nb_lshift*/
	0,			/*nb_rshift*/
	0,			/*nb_and*/
	0,			/*nb_xor*/
	0,			/*nb_or*/
	0,			/*nb_coerce*/
	0,			/*nb_int*/
	0,			/*nb_long*/
	0,			/*nb_float*/
	0,			/*nb_oct*/
	0, 	 	 	/*nb_hex*/
	0,			/*nb_inplace_add*/
	0,			/*nb_inplace_subtract*/
	0,			/*nb_inplace_multiply*/
	0,			/*nb_inplace_divide*/
	0,			/*nb_inplace_remainder*/
	0,			/*nb_inplace_power*/
	0,			/*nb_inplace_lshift*/
	0,			/*nb_inplace_rshift*/
	0,			/*nb_inplace_and*/
	0,			/*nb_inplace_xor*/
	0,			/*nb_inplace_or*/
	0,			/*nb_floor_divide */
	0,			/*nb_true_divide */
	0,			/*nb_inplace_floor_divide */
	0,			/*nb_inplace_true_divide */
};
#define comm_as_number   &comm_number_methods
#define comm_as_sequence (PySequenceMethods*)0
#define comm_as_mapping  (PyMappingMethods*)0


#define comm_methods 0
#define comm_members 0
#define comm_getset 0


PyDoc_STRVAR(comm_doc, "Comm type");

static PyTypeObject PyMPIComm_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,					/*ob_size*/
  "libmpi.Comm",			/*tp_name*/
  sizeof(PyMPICommObject),		/*tp_basicsize*/
  0,					/*tp_itemsize*/

  (destructor)comm_dealloc,		/*tp_dealloc*/
  (printfunc)0,				/*tp_print*/
  (getattrfunc)0,			/*tp_getattr*/
  (setattrfunc)0,			/*tp_setattr*/
  (cmpfunc)0,				/*tp_compare*/
  (reprfunc)0,				/*tp_repr*/

  comm_as_number,			/*tp_as_number*/
  comm_as_sequence,			/*tp_as_sequence*/
  comm_as_mapping,			/*tp_as_mapping*/

  (hashfunc)0,				/*tp_hash*/
  (ternaryfunc)0,			/*tp_call*/
  (reprfunc)0,				/*tp_str*/
  (getattrofunc)0,			/*tp_getattro*/
  (setattrofunc)0,			/*tp_setattro*/

  (PyBufferProcs*)0,			/*tp_as_buffer*/

  Py_TPFLAGS_DEFAULT  |
  Py_TPFLAGS_BASETYPE,			/*tp_flags*/

  comm_doc, 				/*tp_doc*/

  (traverseproc)0,			/* tp_traverse */

  (inquiry)0,				/* tp_clear */

  (richcmpfunc)comm_richcompare,	/* tp_richcompare */

  (long)0,				/* tp_weaklistoffset */

  (getiterfunc)0,			/* tp_iter */
  (iternextfunc)0,			/* tp_iternext */

  comm_methods,			/* tp_methods */
  comm_members,			/* tp_members */
  comm_getset,				/* tp_getset */
  0,					/* tp_base */
  0,					/* tp_dict */
  (descrgetfunc)0,			/* tp_descr_get */
  (descrsetfunc)0,			/* tp_descr_set */
  (long)0,				/* tp_dictoffset */

  comm_init,				/* tp_init */
  comm_alloc,				/* tp_alloc */
  comm_new,				/* tp_new */
  PyObject_Del,           		/* tp_free */
};



/* PyMPIComm C API */

static int
PyMPIComm_Check(PyObject *op)
{
  return PyObject_TypeCheck(op, &PyMPIComm_Type);
}

static int
PyMPIComm_CheckExact(PyObject *op)
{
  return ((op)->ob_type == &PyMPIComm_Type);
}

static PyObject*
PyMPIComm_FromComm(MPI_Comm comm)
{
  PyMPICommObject *self;
  PyTypeObject *type = &PyMPIComm_Type;
  self = (PyMPICommObject*) type->tp_alloc(type, 0);
  if (self != NULL) {
    self->comm = (PyObject*) PyObject_New(_Py_MPI_Comm,
					   &_Py_MPI_Comm_Type);
    if (self->comm == NULL) {
      Py_DECREF(self);
      return NULL;
    }
    PyMPIComm_AS_COMM(self) = comm;
  }
  return (PyObject*)self;
}

static MPI_Comm
PyMPIComm_AsComm(PyObject *self)
{
  if (self) {
    if (PyMPIComm_Check(self))
      return PyMPIComm_AS_COMM(self);
    PyErr_SetString(PyExc_TypeError,
		    "argument must be a 'MPI_Comm' object");
  }
  if (!PyErr_Occurred())
    PyErr_SetString(PyExc_TypeError,
		    "PyMPIComm_AsComm() called with null pointer");
  return MPI_COMM_DEFAULT;
}

static MPI_Comm*
PyMPIComm_AsCommPtr(PyObject *self)
{
  if (self) {
    if (PyMPIComm_Check(self))
      return PyMPIComm_AS_COMM_PTR(self);
    PyErr_SetString(PyExc_TypeError,
		    "argument must be a 'MPI_Comm' object");
  }
  if (!PyErr_Occurred())
    PyErr_SetString(PyExc_TypeError,
		    "PyMPIComm_AsCommPtr() called with null pointer");
  return NULL;
}


/* API pointers to export */

static PyTypeObject* PyMPIComm_TYPE = &PyMPIComm_Type;

static void* PyMPIComm_API[] = {
  /*[0]*/ (void*) PyMPIComm_Check,
  /*[1]*/ (void*) PyMPIComm_CheckExact,
  /*[2]*/ (void*) PyMPIComm_FromComm,
  /*[3]*/ (void*) PyMPIComm_AsComm,
  /*[4]*/ (void*) PyMPIComm_AsCommPtr,
};

/* PyMPIDatatype_Type implementation */

static PyTypeObject _Py_MPI_Datatype_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,				/*ob_size*/
  "_Py_MPI_Datatype",		/*tp_name*/
  sizeof(_Py_MPI_Datatype),	/*tp_basicsize*/
  0,				/*tp_itemsize*/
  (destructor)PyObject_Del,	/*tp_dealloc*/
  0,
};


#define MPI_DATATYPE_DEFAULT MPI_DATATYPE_NULL

#define datatype_alloc PyType_GenericAlloc

static void
datatype_dealloc(PyMPIDatatypeObject *self)
{
  Py_XDECREF(self->datatype);
  self->ob_type->tp_free(self);
}

static PyObject *
datatype_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
  PyObject *ob = NULL;
  static char *kwlist[] = {"datatype", 0};
  
  if (!PyArg_ParseTupleAndKeywords(args, kwds, 
				   "|O:Datatype", kwlist,
				   &ob)) return NULL;
  
  if (ob == NULL || ob == Py_None) {
    PyMPIDatatypeObject *self;
    self = (PyMPIDatatypeObject*) type->tp_alloc(type, 0);
    if (self != NULL) {
      self->datatype = (PyObject*) PyObject_New(_Py_MPI_Datatype,
					     &_Py_MPI_Datatype_Type);
      if (self->datatype == NULL) {
	Py_DECREF(self);
	return NULL;
      }
      PyMPIDatatype_AS_DATATYPE(self) = MPI_DATATYPE_DEFAULT;
    }
    return (PyObject*)self;
  }
  
  if (ob->ob_type == type) {
    Py_INCREF(ob);
    return ob;
  }

  if (PyMPIDatatype_Check(ob)) {
    PyMPIDatatypeObject *self;
    self = (PyMPIDatatypeObject*) type->tp_alloc(type, 0);
    if (self != NULL) {
      self->datatype = ((PyMPIDatatypeObject*)ob)->datatype;
      if (self->datatype == NULL) {
	Py_DECREF(self);
	PyErr_BadInternalCall();
	return NULL;
      }
      Py_INCREF(self->datatype);
    }
    return (PyObject*) self;
  }
  
  PyErr_SetString(PyExc_TypeError,
		  "Datatype() argument must be a 'MPI_Datatype' object");
  return NULL;
}

#define datatype_init (initproc)0

static PyObject *
datatype_richcompare(PyMPIDatatypeObject * o1, PyMPIDatatypeObject * o2, int op)
{
  PyObject *res;

  /* Make sure both arguments are the same type */
  if (!(PyMPIDatatype_Check((PyObject*)o1) && 
	PyMPIDatatype_Check((PyObject*)o2))) {
    Py_INCREF(Py_NotImplemented);
    return Py_NotImplemented;
  }
  if (op != Py_EQ && op != Py_NE) {
    PyErr_SetString(PyExc_TypeError,
		    "cannot compare 'MPI_Datatype' objects using <, <=, >, >=");
    return NULL;
  }

  if ((op == Py_EQ) == 
      (PyMPIDatatype_AS_DATATYPE(o1) == PyMPIDatatype_AS_DATATYPE(o2)))
    res = Py_True;
  else
    res = Py_False;
  
  Py_INCREF(res);
  return res;
}


static int
datatype_nonzero(PyMPIDatatypeObject* ob) {
  return PyMPIDatatype_AS_DATATYPE(ob) != MPI_DATATYPE_NULL;
}
static PyNumberMethods datatype_number_methods = {
	0,			/*nb_add*/
	0,			/*nb_subtract*/
	0,			/*nb_multiply*/
	0, 	 	 	/*nb_divide*/
	0,			/*nb_remainder*/
	0,			/*nb_divmod*/
	0,			/*nb_power*/
	0,			/*nb_negative*/
	0,			/*nb_positive*/
	0,			/*nb_absolute*/
	(inquiry)datatype_nonzero,	/*nb_nonzero*/
	0,			/*nb_invert*/
	0,			/*nb_lshift*/
	0,			/*nb_rshift*/
	0,			/*nb_and*/
	0,			/*nb_xor*/
	0,			/*nb_or*/
	0,			/*nb_coerce*/
	0,			/*nb_int*/
	0,			/*nb_long*/
	0,			/*nb_float*/
	0,			/*nb_oct*/
	0, 	 	 	/*nb_hex*/
	0,			/*nb_inplace_add*/
	0,			/*nb_inplace_subtract*/
	0,			/*nb_inplace_multiply*/
	0,			/*nb_inplace_divide*/
	0,			/*nb_inplace_remainder*/
	0,			/*nb_inplace_power*/
	0,			/*nb_inplace_lshift*/
	0,			/*nb_inplace_rshift*/
	0,			/*nb_inplace_and*/
	0,			/*nb_inplace_xor*/
	0,			/*nb_inplace_or*/
	0,			/*nb_floor_divide */
	0,			/*nb_true_divide */
	0,			/*nb_inplace_floor_divide */
	0,			/*nb_inplace_true_divide */
};
#define datatype_as_number   &datatype_number_methods
#define datatype_as_sequence (PySequenceMethods*)0
#define datatype_as_mapping  (PyMappingMethods*)0


#define datatype_methods 0
#define datatype_members 0
#define datatype_getset 0


PyDoc_STRVAR(datatype_doc, "Datatype type");

static PyTypeObject PyMPIDatatype_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,					/*ob_size*/
  "libmpi.Datatype",			/*tp_name*/
  sizeof(PyMPIDatatypeObject),		/*tp_basicsize*/
  0,					/*tp_itemsize*/

  (destructor)datatype_dealloc,		/*tp_dealloc*/
  (printfunc)0,				/*tp_print*/
  (getattrfunc)0,			/*tp_getattr*/
  (setattrfunc)0,			/*tp_setattr*/
  (cmpfunc)0,				/*tp_compare*/
  (reprfunc)0,				/*tp_repr*/

  datatype_as_number,			/*tp_as_number*/
  datatype_as_sequence,			/*tp_as_sequence*/
  datatype_as_mapping,			/*tp_as_mapping*/

  (hashfunc)0,				/*tp_hash*/
  (ternaryfunc)0,			/*tp_call*/
  (reprfunc)0,				/*tp_str*/
  (getattrofunc)0,			/*tp_getattro*/
  (setattrofunc)0,			/*tp_setattro*/

  (PyBufferProcs*)0,			/*tp_as_buffer*/

  Py_TPFLAGS_DEFAULT  |
  Py_TPFLAGS_BASETYPE,			/*tp_flags*/

  datatype_doc, 				/*tp_doc*/

  (traverseproc)0,			/* tp_traverse */

  (inquiry)0,				/* tp_clear */

  (richcmpfunc)datatype_richcompare,	/* tp_richcompare */

  (long)0,				/* tp_weaklistoffset */

  (getiterfunc)0,			/* tp_iter */
  (iternextfunc)0,			/* tp_iternext */

  datatype_methods,			/* tp_methods */
  datatype_members,			/* tp_members */
  datatype_getset,				/* tp_getset */
  0,					/* tp_base */
  0,					/* tp_dict */
  (descrgetfunc)0,			/* tp_descr_get */
  (descrsetfunc)0,			/* tp_descr_set */
  (long)0,				/* tp_dictoffset */

  datatype_init,				/* tp_init */
  datatype_alloc,				/* tp_alloc */
  datatype_new,				/* tp_new */
  PyObject_Del,           		/* tp_free */
};



/* PyMPIDatatype C API */

static int
PyMPIDatatype_Check(PyObject *op)
{
  return PyObject_TypeCheck(op, &PyMPIDatatype_Type);
}

static int
PyMPIDatatype_CheckExact(PyObject *op)
{
  return ((op)->ob_type == &PyMPIDatatype_Type);
}

static PyObject*
PyMPIDatatype_FromDatatype(MPI_Datatype datatype)
{
  PyMPIDatatypeObject *self;
  PyTypeObject *type = &PyMPIDatatype_Type;
  self = (PyMPIDatatypeObject*) type->tp_alloc(type, 0);
  if (self != NULL) {
    self->datatype = (PyObject*) PyObject_New(_Py_MPI_Datatype,
					   &_Py_MPI_Datatype_Type);
    if (self->datatype == NULL) {
      Py_DECREF(self);
      return NULL;
    }
    PyMPIDatatype_AS_DATATYPE(self) = datatype;
  }
  return (PyObject*)self;
}

static MPI_Datatype
PyMPIDatatype_AsDatatype(PyObject *self)
{
  if (self) {
    if (PyMPIDatatype_Check(self))
      return PyMPIDatatype_AS_DATATYPE(self);
    PyErr_SetString(PyExc_TypeError,
		    "argument must be a 'MPI_Datatype' object");
  }
  if (!PyErr_Occurred())
    PyErr_SetString(PyExc_TypeError,
		    "PyMPIDatatype_AsDatatype() called with null pointer");
  return MPI_DATATYPE_DEFAULT;
}

static MPI_Datatype*
PyMPIDatatype_AsDatatypePtr(PyObject *self)
{
  if (self) {
    if (PyMPIDatatype_Check(self))
      return PyMPIDatatype_AS_DATATYPE_PTR(self);
    PyErr_SetString(PyExc_TypeError,
		    "argument must be a 'MPI_Datatype' object");
  }
  if (!PyErr_Occurred())
    PyErr_SetString(PyExc_TypeError,
		    "PyMPIDatatype_AsDatatypePtr() called with null pointer");
  return NULL;
}


/* API pointers to export */

static PyTypeObject* PyMPIDatatype_TYPE = &PyMPIDatatype_Type;

static void* PyMPIDatatype_API[] = {
  /*[0]*/ (void*) PyMPIDatatype_Check,
  /*[1]*/ (void*) PyMPIDatatype_CheckExact,
  /*[2]*/ (void*) PyMPIDatatype_FromDatatype,
  /*[3]*/ (void*) PyMPIDatatype_AsDatatype,
  /*[4]*/ (void*) PyMPIDatatype_AsDatatypePtr,
};

/* PyMPIErrhandler_Type implementation */

static PyTypeObject _Py_MPI_Errhandler_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,				/*ob_size*/
  "_Py_MPI_Errhandler",		/*tp_name*/
  sizeof(_Py_MPI_Errhandler),	/*tp_basicsize*/
  0,				/*tp_itemsize*/
  (destructor)PyObject_Del,	/*tp_dealloc*/
  0,
};


#define MPI_ERRHANDLER_DEFAULT MPI_ERRHANDLER_NULL

#define errhandler_alloc PyType_GenericAlloc

static void
errhandler_dealloc(PyMPIErrhandlerObject *self)
{
  Py_XDECREF(self->errhandler);
  self->ob_type->tp_free(self);
}

static PyObject *
errhandler_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
  PyObject *ob = NULL;
  static char *kwlist[] = {"errhandler", 0};
  
  if (!PyArg_ParseTupleAndKeywords(args, kwds, 
				   "|O:Errhandler", kwlist,
				   &ob)) return NULL;
  
  if (ob == NULL || ob == Py_None) {
    PyMPIErrhandlerObject *self;
    self = (PyMPIErrhandlerObject*) type->tp_alloc(type, 0);
    if (self != NULL) {
      self->errhandler = (PyObject*) PyObject_New(_Py_MPI_Errhandler,
					     &_Py_MPI_Errhandler_Type);
      if (self->errhandler == NULL) {
	Py_DECREF(self);
	return NULL;
      }
      PyMPIErrhandler_AS_ERRHANDLER(self) = MPI_ERRHANDLER_DEFAULT;
    }
    return (PyObject*)self;
  }
  
  if (ob->ob_type == type) {
    Py_INCREF(ob);
    return ob;
  }

  if (PyMPIErrhandler_Check(ob)) {
    PyMPIErrhandlerObject *self;
    self = (PyMPIErrhandlerObject*) type->tp_alloc(type, 0);
    if (self != NULL) {
      self->errhandler = ((PyMPIErrhandlerObject*)ob)->errhandler;
      if (self->errhandler == NULL) {
	Py_DECREF(self);
	PyErr_BadInternalCall();
	return NULL;
      }
      Py_INCREF(self->errhandler);
    }
    return (PyObject*) self;
  }
  
  PyErr_SetString(PyExc_TypeError,
		  "Errhandler() argument must be a 'MPI_Errhandler' object");
  return NULL;
}

#define errhandler_init (initproc)0

static PyObject *
errhandler_richcompare(PyMPIErrhandlerObject * o1, PyMPIErrhandlerObject * o2, int op)
{
  PyObject *res;

  /* Make sure both arguments are the same type */
  if (!(PyMPIErrhandler_Check((PyObject*)o1) && 
	PyMPIErrhandler_Check((PyObject*)o2))) {
    Py_INCREF(Py_NotImplemented);
    return Py_NotImplemented;
  }
  if (op != Py_EQ && op != Py_NE) {
    PyErr_SetString(PyExc_TypeError,
		    "cannot compare 'MPI_Errhandler' objects using <, <=, >, >=");
    return NULL;
  }

  if ((op == Py_EQ) == 
      (PyMPIErrhandler_AS_ERRHANDLER(o1) == PyMPIErrhandler_AS_ERRHANDLER(o2)))
    res = Py_True;
  else
    res = Py_False;
  
  Py_INCREF(res);
  return res;
}


static int
errhandler_nonzero(PyMPIErrhandlerObject* ob) {
  return PyMPIErrhandler_AS_ERRHANDLER(ob) != MPI_ERRHANDLER_NULL;
}
static PyNumberMethods errhandler_number_methods = {
	0,			/*nb_add*/
	0,			/*nb_subtract*/
	0,			/*nb_multiply*/
	0, 	 	 	/*nb_divide*/
	0,			/*nb_remainder*/
	0,			/*nb_divmod*/
	0,			/*nb_power*/
	0,			/*nb_negative*/
	0,			/*nb_positive*/
	0,			/*nb_absolute*/
	(inquiry)errhandler_nonzero,	/*nb_nonzero*/
	0,			/*nb_invert*/
	0,			/*nb_lshift*/
	0,			/*nb_rshift*/
	0,			/*nb_and*/
	0,			/*nb_xor*/
	0,			/*nb_or*/
	0,			/*nb_coerce*/
	0,			/*nb_int*/
	0,			/*nb_long*/
	0,			/*nb_float*/
	0,			/*nb_oct*/
	0, 	 	 	/*nb_hex*/
	0,			/*nb_inplace_add*/
	0,			/*nb_inplace_subtract*/
	0,			/*nb_inplace_multiply*/
	0,			/*nb_inplace_divide*/
	0,			/*nb_inplace_remainder*/
	0,			/*nb_inplace_power*/
	0,			/*nb_inplace_lshift*/
	0,			/*nb_inplace_rshift*/
	0,			/*nb_inplace_and*/
	0,			/*nb_inplace_xor*/
	0,			/*nb_inplace_or*/
	0,			/*nb_floor_divide */
	0,			/*nb_true_divide */
	0,			/*nb_inplace_floor_divide */
	0,			/*nb_inplace_true_divide */
};
#define errhandler_as_number   &errhandler_number_methods
#define errhandler_as_sequence (PySequenceMethods*)0
#define errhandler_as_mapping  (PyMappingMethods*)0


#define errhandler_methods 0
#define errhandler_members 0
#define errhandler_getset 0


PyDoc_STRVAR(errhandler_doc, "Errhandler type");

static PyTypeObject PyMPIErrhandler_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,					/*ob_size*/
  "libmpi.Errhandler",			/*tp_name*/
  sizeof(PyMPIErrhandlerObject),		/*tp_basicsize*/
  0,					/*tp_itemsize*/

  (destructor)errhandler_dealloc,		/*tp_dealloc*/
  (printfunc)0,				/*tp_print*/
  (getattrfunc)0,			/*tp_getattr*/
  (setattrfunc)0,			/*tp_setattr*/
  (cmpfunc)0,				/*tp_compare*/
  (reprfunc)0,				/*tp_repr*/

  errhandler_as_number,			/*tp_as_number*/
  errhandler_as_sequence,			/*tp_as_sequence*/
  errhandler_as_mapping,			/*tp_as_mapping*/

  (hashfunc)0,				/*tp_hash*/
  (ternaryfunc)0,			/*tp_call*/
  (reprfunc)0,				/*tp_str*/
  (getattrofunc)0,			/*tp_getattro*/
  (setattrofunc)0,			/*tp_setattro*/

  (PyBufferProcs*)0,			/*tp_as_buffer*/

  Py_TPFLAGS_DEFAULT  |
  Py_TPFLAGS_BASETYPE,			/*tp_flags*/

  errhandler_doc, 				/*tp_doc*/

  (traverseproc)0,			/* tp_traverse */

  (inquiry)0,				/* tp_clear */

  (richcmpfunc)errhandler_richcompare,	/* tp_richcompare */

  (long)0,				/* tp_weaklistoffset */

  (getiterfunc)0,			/* tp_iter */
  (iternextfunc)0,			/* tp_iternext */

  errhandler_methods,			/* tp_methods */
  errhandler_members,			/* tp_members */
  errhandler_getset,				/* tp_getset */
  0,					/* tp_base */
  0,					/* tp_dict */
  (descrgetfunc)0,			/* tp_descr_get */
  (descrsetfunc)0,			/* tp_descr_set */
  (long)0,				/* tp_dictoffset */

  errhandler_init,				/* tp_init */
  errhandler_alloc,				/* tp_alloc */
  errhandler_new,				/* tp_new */
  PyObject_Del,           		/* tp_free */
};



/* PyMPIErrhandler C API */

static int
PyMPIErrhandler_Check(PyObject *op)
{
  return PyObject_TypeCheck(op, &PyMPIErrhandler_Type);
}

static int
PyMPIErrhandler_CheckExact(PyObject *op)
{
  return ((op)->ob_type == &PyMPIErrhandler_Type);
}

static PyObject*
PyMPIErrhandler_FromErrhandler(MPI_Errhandler errhandler)
{
  PyMPIErrhandlerObject *self;
  PyTypeObject *type = &PyMPIErrhandler_Type;
  self = (PyMPIErrhandlerObject*) type->tp_alloc(type, 0);
  if (self != NULL) {
    self->errhandler = (PyObject*) PyObject_New(_Py_MPI_Errhandler,
					   &_Py_MPI_Errhandler_Type);
    if (self->errhandler == NULL) {
      Py_DECREF(self);
      return NULL;
    }
    PyMPIErrhandler_AS_ERRHANDLER(self) = errhandler;
  }
  return (PyObject*)self;
}

static MPI_Errhandler
PyMPIErrhandler_AsErrhandler(PyObject *self)
{
  if (self) {
    if (PyMPIErrhandler_Check(self))
      return PyMPIErrhandler_AS_ERRHANDLER(self);
    PyErr_SetString(PyExc_TypeError,
		    "argument must be a 'MPI_Errhandler' object");
  }
  if (!PyErr_Occurred())
    PyErr_SetString(PyExc_TypeError,
		    "PyMPIErrhandler_AsErrhandler() called with null pointer");
  return MPI_ERRHANDLER_DEFAULT;
}

static MPI_Errhandler*
PyMPIErrhandler_AsErrhandlerPtr(PyObject *self)
{
  if (self) {
    if (PyMPIErrhandler_Check(self))
      return PyMPIErrhandler_AS_ERRHANDLER_PTR(self);
    PyErr_SetString(PyExc_TypeError,
		    "argument must be a 'MPI_Errhandler' object");
  }
  if (!PyErr_Occurred())
    PyErr_SetString(PyExc_TypeError,
		    "PyMPIErrhandler_AsErrhandlerPtr() called with null pointer");
  return NULL;
}


/* API pointers to export */

static PyTypeObject* PyMPIErrhandler_TYPE = &PyMPIErrhandler_Type;

static void* PyMPIErrhandler_API[] = {
  /*[0]*/ (void*) PyMPIErrhandler_Check,
  /*[1]*/ (void*) PyMPIErrhandler_CheckExact,
  /*[2]*/ (void*) PyMPIErrhandler_FromErrhandler,
  /*[3]*/ (void*) PyMPIErrhandler_AsErrhandler,
  /*[4]*/ (void*) PyMPIErrhandler_AsErrhandlerPtr,
};

/* PyMPIGroup_Type implementation */

static PyTypeObject _Py_MPI_Group_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,				/*ob_size*/
  "_Py_MPI_Group",		/*tp_name*/
  sizeof(_Py_MPI_Group),	/*tp_basicsize*/
  0,				/*tp_itemsize*/
  (destructor)PyObject_Del,	/*tp_dealloc*/
  0,
};


#define MPI_GROUP_DEFAULT MPI_GROUP_NULL

#define group_alloc PyType_GenericAlloc

static void
group_dealloc(PyMPIGroupObject *self)
{
  Py_XDECREF(self->group);
  self->ob_type->tp_free(self);
}

static PyObject *
group_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
  PyObject *ob = NULL;
  static char *kwlist[] = {"group", 0};
  
  if (!PyArg_ParseTupleAndKeywords(args, kwds, 
				   "|O:Group", kwlist,
				   &ob)) return NULL;
  
  if (ob == NULL || ob == Py_None) {
    PyMPIGroupObject *self;
    self = (PyMPIGroupObject*) type->tp_alloc(type, 0);
    if (self != NULL) {
      self->group = (PyObject*) PyObject_New(_Py_MPI_Group,
					     &_Py_MPI_Group_Type);
      if (self->group == NULL) {
	Py_DECREF(self);
	return NULL;
      }
      PyMPIGroup_AS_GROUP(self) = MPI_GROUP_DEFAULT;
    }
    return (PyObject*)self;
  }
  
  if (ob->ob_type == type) {
    Py_INCREF(ob);
    return ob;
  }

  if (PyMPIGroup_Check(ob)) {
    PyMPIGroupObject *self;
    self = (PyMPIGroupObject*) type->tp_alloc(type, 0);
    if (self != NULL) {
      self->group = ((PyMPIGroupObject*)ob)->group;
      if (self->group == NULL) {
	Py_DECREF(self);
	PyErr_BadInternalCall();
	return NULL;
      }
      Py_INCREF(self->group);
    }
    return (PyObject*) self;
  }
  
  PyErr_SetString(PyExc_TypeError,
		  "Group() argument must be a 'MPI_Group' object");
  return NULL;
}

#define group_init (initproc)0

static PyObject *
group_richcompare(PyMPIGroupObject * o1, PyMPIGroupObject * o2, int op)
{
  PyObject *res;

  /* Make sure both arguments are the same type */
  if (!(PyMPIGroup_Check((PyObject*)o1) && 
	PyMPIGroup_Check((PyObject*)o2))) {
    Py_INCREF(Py_NotImplemented);
    return Py_NotImplemented;
  }
  if (op != Py_EQ && op != Py_NE) {
    PyErr_SetString(PyExc_TypeError,
		    "cannot compare 'MPI_Group' objects using <, <=, >, >=");
    return NULL;
  }

  if ((op == Py_EQ) == 
      (PyMPIGroup_AS_GROUP(o1) == PyMPIGroup_AS_GROUP(o2)))
    res = Py_True;
  else
    res = Py_False;
  
  Py_INCREF(res);
  return res;
}


static int
group_nonzero(PyMPIGroupObject* ob) {
  return PyMPIGroup_AS_GROUP(ob) != MPI_GROUP_NULL;
}
static PyNumberMethods group_number_methods = {
	0,			/*nb_add*/
	0,			/*nb_subtract*/
	0,			/*nb_multiply*/
	0, 	 	 	/*nb_divide*/
	0,			/*nb_remainder*/
	0,			/*nb_divmod*/
	0,			/*nb_power*/
	0,			/*nb_negative*/
	0,			/*nb_positive*/
	0,			/*nb_absolute*/
	(inquiry)group_nonzero,	/*nb_nonzero*/
	0,			/*nb_invert*/
	0,			/*nb_lshift*/
	0,			/*nb_rshift*/
	0,			/*nb_and*/
	0,			/*nb_xor*/
	0,			/*nb_or*/
	0,			/*nb_coerce*/
	0,			/*nb_int*/
	0,			/*nb_long*/
	0,			/*nb_float*/
	0,			/*nb_oct*/
	0, 	 	 	/*nb_hex*/
	0,			/*nb_inplace_add*/
	0,			/*nb_inplace_subtract*/
	0,			/*nb_inplace_multiply*/
	0,			/*nb_inplace_divide*/
	0,			/*nb_inplace_remainder*/
	0,			/*nb_inplace_power*/
	0,			/*nb_inplace_lshift*/
	0,			/*nb_inplace_rshift*/
	0,			/*nb_inplace_and*/
	0,			/*nb_inplace_xor*/
	0,			/*nb_inplace_or*/
	0,			/*nb_floor_divide */
	0,			/*nb_true_divide */
	0,			/*nb_inplace_floor_divide */
	0,			/*nb_inplace_true_divide */
};
#define group_as_number   &group_number_methods
#define group_as_sequence (PySequenceMethods*)0
#define group_as_mapping  (PyMappingMethods*)0


#define group_methods 0
#define group_members 0
#define group_getset 0


PyDoc_STRVAR(group_doc, "Group type");

static PyTypeObject PyMPIGroup_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,					/*ob_size*/
  "libmpi.Group",			/*tp_name*/
  sizeof(PyMPIGroupObject),		/*tp_basicsize*/
  0,					/*tp_itemsize*/

  (destructor)group_dealloc,		/*tp_dealloc*/
  (printfunc)0,				/*tp_print*/
  (getattrfunc)0,			/*tp_getattr*/
  (setattrfunc)0,			/*tp_setattr*/
  (cmpfunc)0,				/*tp_compare*/
  (reprfunc)0,				/*tp_repr*/

  group_as_number,			/*tp_as_number*/
  group_as_sequence,			/*tp_as_sequence*/
  group_as_mapping,			/*tp_as_mapping*/

  (hashfunc)0,				/*tp_hash*/
  (ternaryfunc)0,			/*tp_call*/
  (reprfunc)0,				/*tp_str*/
  (getattrofunc)0,			/*tp_getattro*/
  (setattrofunc)0,			/*tp_setattro*/

  (PyBufferProcs*)0,			/*tp_as_buffer*/

  Py_TPFLAGS_DEFAULT  |
  Py_TPFLAGS_BASETYPE,			/*tp_flags*/

  group_doc, 				/*tp_doc*/

  (traverseproc)0,			/* tp_traverse */

  (inquiry)0,				/* tp_clear */

  (richcmpfunc)group_richcompare,	/* tp_richcompare */

  (long)0,				/* tp_weaklistoffset */

  (getiterfunc)0,			/* tp_iter */
  (iternextfunc)0,			/* tp_iternext */

  group_methods,			/* tp_methods */
  group_members,			/* tp_members */
  group_getset,				/* tp_getset */
  0,					/* tp_base */
  0,					/* tp_dict */
  (descrgetfunc)0,			/* tp_descr_get */
  (descrsetfunc)0,			/* tp_descr_set */
  (long)0,				/* tp_dictoffset */

  group_init,				/* tp_init */
  group_alloc,				/* tp_alloc */
  group_new,				/* tp_new */
  PyObject_Del,           		/* tp_free */
};



/* PyMPIGroup C API */

static int
PyMPIGroup_Check(PyObject *op)
{
  return PyObject_TypeCheck(op, &PyMPIGroup_Type);
}

static int
PyMPIGroup_CheckExact(PyObject *op)
{
  return ((op)->ob_type == &PyMPIGroup_Type);
}

static PyObject*
PyMPIGroup_FromGroup(MPI_Group group)
{
  PyMPIGroupObject *self;
  PyTypeObject *type = &PyMPIGroup_Type;
  self = (PyMPIGroupObject*) type->tp_alloc(type, 0);
  if (self != NULL) {
    self->group = (PyObject*) PyObject_New(_Py_MPI_Group,
					   &_Py_MPI_Group_Type);
    if (self->group == NULL) {
      Py_DECREF(self);
      return NULL;
    }
    PyMPIGroup_AS_GROUP(self) = group;
  }
  return (PyObject*)self;
}

static MPI_Group
PyMPIGroup_AsGroup(PyObject *self)
{
  if (self) {
    if (PyMPIGroup_Check(self))
      return PyMPIGroup_AS_GROUP(self);
    PyErr_SetString(PyExc_TypeError,
		    "argument must be a 'MPI_Group' object");
  }
  if (!PyErr_Occurred())
    PyErr_SetString(PyExc_TypeError,
		    "PyMPIGroup_AsGroup() called with null pointer");
  return MPI_GROUP_DEFAULT;
}

static MPI_Group*
PyMPIGroup_AsGroupPtr(PyObject *self)
{
  if (self) {
    if (PyMPIGroup_Check(self))
      return PyMPIGroup_AS_GROUP_PTR(self);
    PyErr_SetString(PyExc_TypeError,
		    "argument must be a 'MPI_Group' object");
  }
  if (!PyErr_Occurred())
    PyErr_SetString(PyExc_TypeError,
		    "PyMPIGroup_AsGroupPtr() called with null pointer");
  return NULL;
}


/* API pointers to export */

static PyTypeObject* PyMPIGroup_TYPE = &PyMPIGroup_Type;

static void* PyMPIGroup_API[] = {
  /*[0]*/ (void*) PyMPIGroup_Check,
  /*[1]*/ (void*) PyMPIGroup_CheckExact,
  /*[2]*/ (void*) PyMPIGroup_FromGroup,
  /*[3]*/ (void*) PyMPIGroup_AsGroup,
  /*[4]*/ (void*) PyMPIGroup_AsGroupPtr,
};

/* PyMPIOp_Type implementation */

static PyTypeObject _Py_MPI_Op_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,				/*ob_size*/
  "_Py_MPI_Op",		/*tp_name*/
  sizeof(_Py_MPI_Op),	/*tp_basicsize*/
  0,				/*tp_itemsize*/
  (destructor)PyObject_Del,	/*tp_dealloc*/
  0,
};


#define MPI_OP_DEFAULT MPI_OP_NULL

#define op_alloc PyType_GenericAlloc

static void
op_dealloc(PyMPIOpObject *self)
{
  Py_XDECREF(self->op);
  self->ob_type->tp_free(self);
}

static PyObject *
op_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
  PyObject *ob = NULL;
  static char *kwlist[] = {"op", 0};
  
  if (!PyArg_ParseTupleAndKeywords(args, kwds, 
				   "|O:Op", kwlist,
				   &ob)) return NULL;
  
  if (ob == NULL || ob == Py_None) {
    PyMPIOpObject *self;
    self = (PyMPIOpObject*) type->tp_alloc(type, 0);
    if (self != NULL) {
      self->op = (PyObject*) PyObject_New(_Py_MPI_Op,
					     &_Py_MPI_Op_Type);
      if (self->op == NULL) {
	Py_DECREF(self);
	return NULL;
      }
      PyMPIOp_AS_OP(self) = MPI_OP_DEFAULT;
    }
    return (PyObject*)self;
  }
  
  if (ob->ob_type == type) {
    Py_INCREF(ob);
    return ob;
  }

  if (PyMPIOp_Check(ob)) {
    PyMPIOpObject *self;
    self = (PyMPIOpObject*) type->tp_alloc(type, 0);
    if (self != NULL) {
      self->op = ((PyMPIOpObject*)ob)->op;
      if (self->op == NULL) {
	Py_DECREF(self);
	PyErr_BadInternalCall();
	return NULL;
      }
      Py_INCREF(self->op);
    }
    return (PyObject*) self;
  }
  
  PyErr_SetString(PyExc_TypeError,
		  "Op() argument must be a 'MPI_Op' object");
  return NULL;
}

#define op_init (initproc)0

static PyObject *
op_richcompare(PyMPIOpObject * o1, PyMPIOpObject * o2, int op)
{
  PyObject *res;

  /* Make sure both arguments are the same type */
  if (!(PyMPIOp_Check((PyObject*)o1) && 
	PyMPIOp_Check((PyObject*)o2))) {
    Py_INCREF(Py_NotImplemented);
    return Py_NotImplemented;
  }
  if (op != Py_EQ && op != Py_NE) {
    PyErr_SetString(PyExc_TypeError,
		    "cannot compare 'MPI_Op' objects using <, <=, >, >=");
    return NULL;
  }

  if ((op == Py_EQ) == 
      (PyMPIOp_AS_OP(o1) == PyMPIOp_AS_OP(o2)))
    res = Py_True;
  else
    res = Py_False;
  
  Py_INCREF(res);
  return res;
}


static int
op_nonzero(PyMPIOpObject* ob) {
  return PyMPIOp_AS_OP(ob) != MPI_OP_NULL;
}
static PyNumberMethods op_number_methods = {
	0,			/*nb_add*/
	0,			/*nb_subtract*/
	0,			/*nb_multiply*/
	0, 	 	 	/*nb_divide*/
	0,			/*nb_remainder*/
	0,			/*nb_divmod*/
	0,			/*nb_power*/
	0,			/*nb_negative*/
	0,			/*nb_positive*/
	0,			/*nb_absolute*/
	(inquiry)op_nonzero,	/*nb_nonzero*/
	0,			/*nb_invert*/
	0,			/*nb_lshift*/
	0,			/*nb_rshift*/
	0,			/*nb_and*/
	0,			/*nb_xor*/
	0,			/*nb_or*/
	0,			/*nb_coerce*/
	0,			/*nb_int*/
	0,			/*nb_long*/
	0,			/*nb_float*/
	0,			/*nb_oct*/
	0, 	 	 	/*nb_hex*/
	0,			/*nb_inplace_add*/
	0,			/*nb_inplace_subtract*/
	0,			/*nb_inplace_multiply*/
	0,			/*nb_inplace_divide*/
	0,			/*nb_inplace_remainder*/
	0,			/*nb_inplace_power*/
	0,			/*nb_inplace_lshift*/
	0,			/*nb_inplace_rshift*/
	0,			/*nb_inplace_and*/
	0,			/*nb_inplace_xor*/
	0,			/*nb_inplace_or*/
	0,			/*nb_floor_divide */
	0,			/*nb_true_divide */
	0,			/*nb_inplace_floor_divide */
	0,			/*nb_inplace_true_divide */
};
#define op_as_number   &op_number_methods
#define op_as_sequence (PySequenceMethods*)0
#define op_as_mapping  (PyMappingMethods*)0


#define op_methods 0
#define op_members 0
#define op_getset 0


PyDoc_STRVAR(op_doc, "Op type");

static PyTypeObject PyMPIOp_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,					/*ob_size*/
  "libmpi.Op",			/*tp_name*/
  sizeof(PyMPIOpObject),		/*tp_basicsize*/
  0,					/*tp_itemsize*/

  (destructor)op_dealloc,		/*tp_dealloc*/
  (printfunc)0,				/*tp_print*/
  (getattrfunc)0,			/*tp_getattr*/
  (setattrfunc)0,			/*tp_setattr*/
  (cmpfunc)0,				/*tp_compare*/
  (reprfunc)0,				/*tp_repr*/

  op_as_number,			/*tp_as_number*/
  op_as_sequence,			/*tp_as_sequence*/
  op_as_mapping,			/*tp_as_mapping*/

  (hashfunc)0,				/*tp_hash*/
  (ternaryfunc)0,			/*tp_call*/
  (reprfunc)0,				/*tp_str*/
  (getattrofunc)0,			/*tp_getattro*/
  (setattrofunc)0,			/*tp_setattro*/

  (PyBufferProcs*)0,			/*tp_as_buffer*/

  Py_TPFLAGS_DEFAULT  |
  Py_TPFLAGS_BASETYPE,			/*tp_flags*/

  op_doc, 				/*tp_doc*/

  (traverseproc)0,			/* tp_traverse */

  (inquiry)0,				/* tp_clear */

  (richcmpfunc)op_richcompare,	/* tp_richcompare */

  (long)0,				/* tp_weaklistoffset */

  (getiterfunc)0,			/* tp_iter */
  (iternextfunc)0,			/* tp_iternext */

  op_methods,			/* tp_methods */
  op_members,			/* tp_members */
  op_getset,				/* tp_getset */
  0,					/* tp_base */
  0,					/* tp_dict */
  (descrgetfunc)0,			/* tp_descr_get */
  (descrsetfunc)0,			/* tp_descr_set */
  (long)0,				/* tp_dictoffset */

  op_init,				/* tp_init */
  op_alloc,				/* tp_alloc */
  op_new,				/* tp_new */
  PyObject_Del,           		/* tp_free */
};



/* PyMPIOp C API */

static int
PyMPIOp_Check(PyObject *op)
{
  return PyObject_TypeCheck(op, &PyMPIOp_Type);
}

static int
PyMPIOp_CheckExact(PyObject *op)
{
  return ((op)->ob_type == &PyMPIOp_Type);
}

static PyObject*
PyMPIOp_FromOp(MPI_Op op)
{
  PyMPIOpObject *self;
  PyTypeObject *type = &PyMPIOp_Type;
  self = (PyMPIOpObject*) type->tp_alloc(type, 0);
  if (self != NULL) {
    self->op = (PyObject*) PyObject_New(_Py_MPI_Op,
					   &_Py_MPI_Op_Type);
    if (self->op == NULL) {
      Py_DECREF(self);
      return NULL;
    }
    PyMPIOp_AS_OP(self) = op;
  }
  return (PyObject*)self;
}

static MPI_Op
PyMPIOp_AsOp(PyObject *self)
{
  if (self) {
    if (PyMPIOp_Check(self))
      return PyMPIOp_AS_OP(self);
    PyErr_SetString(PyExc_TypeError,
		    "argument must be a 'MPI_Op' object");
  }
  if (!PyErr_Occurred())
    PyErr_SetString(PyExc_TypeError,
		    "PyMPIOp_AsOp() called with null pointer");
  return MPI_OP_DEFAULT;
}

static MPI_Op*
PyMPIOp_AsOpPtr(PyObject *self)
{
  if (self) {
    if (PyMPIOp_Check(self))
      return PyMPIOp_AS_OP_PTR(self);
    PyErr_SetString(PyExc_TypeError,
		    "argument must be a 'MPI_Op' object");
  }
  if (!PyErr_Occurred())
    PyErr_SetString(PyExc_TypeError,
		    "PyMPIOp_AsOpPtr() called with null pointer");
  return NULL;
}


/* API pointers to export */

static PyTypeObject* PyMPIOp_TYPE = &PyMPIOp_Type;

static void* PyMPIOp_API[] = {
  /*[0]*/ (void*) PyMPIOp_Check,
  /*[1]*/ (void*) PyMPIOp_CheckExact,
  /*[2]*/ (void*) PyMPIOp_FromOp,
  /*[3]*/ (void*) PyMPIOp_AsOp,
  /*[4]*/ (void*) PyMPIOp_AsOpPtr,
};

/* PyMPIRequest_Type implementation */

static PyTypeObject _Py_MPI_Request_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,				/*ob_size*/
  "_Py_MPI_Request",		/*tp_name*/
  sizeof(_Py_MPI_Request),	/*tp_basicsize*/
  0,				/*tp_itemsize*/
  (destructor)PyObject_Del,	/*tp_dealloc*/
  0,
};


#define MPI_REQUEST_DEFAULT MPI_REQUEST_NULL

#define request_alloc PyType_GenericAlloc

static void
request_dealloc(PyMPIRequestObject *self)
{
  Py_XDECREF(self->request);
  self->ob_type->tp_free(self);
}

static PyObject *
request_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
  PyObject *ob = NULL;
  static char *kwlist[] = {"request", 0};
  
  if (!PyArg_ParseTupleAndKeywords(args, kwds, 
				   "|O:Request", kwlist,
				   &ob)) return NULL;
  
  if (ob == NULL || ob == Py_None) {
    PyMPIRequestObject *self;
    self = (PyMPIRequestObject*) type->tp_alloc(type, 0);
    if (self != NULL) {
      self->request = (PyObject*) PyObject_New(_Py_MPI_Request,
					     &_Py_MPI_Request_Type);
      if (self->request == NULL) {
	Py_DECREF(self);
	return NULL;
      }
      PyMPIRequest_AS_REQUEST(self) = MPI_REQUEST_DEFAULT;
    }
    return (PyObject*)self;
  }
  
  if (ob->ob_type == type) {
    Py_INCREF(ob);
    return ob;
  }

  if (PyMPIRequest_Check(ob)) {
    PyMPIRequestObject *self;
    self = (PyMPIRequestObject*) type->tp_alloc(type, 0);
    if (self != NULL) {
      self->request = ((PyMPIRequestObject*)ob)->request;
      if (self->request == NULL) {
	Py_DECREF(self);
	PyErr_BadInternalCall();
	return NULL;
      }
      Py_INCREF(self->request);
    }
    return (PyObject*) self;
  }
  
  PyErr_SetString(PyExc_TypeError,
		  "Request() argument must be a 'MPI_Request' object");
  return NULL;
}

#define request_init (initproc)0

static PyObject *
request_richcompare(PyMPIRequestObject * o1, PyMPIRequestObject * o2, int op)
{
  PyObject *res;

  /* Make sure both arguments are the same type */
  if (!(PyMPIRequest_Check((PyObject*)o1) && 
	PyMPIRequest_Check((PyObject*)o2))) {
    Py_INCREF(Py_NotImplemented);
    return Py_NotImplemented;
  }
  if (op != Py_EQ && op != Py_NE) {
    PyErr_SetString(PyExc_TypeError,
		    "cannot compare 'MPI_Request' objects using <, <=, >, >=");
    return NULL;
  }

  if ((op == Py_EQ) == 
      (PyMPIRequest_AS_REQUEST(o1) == PyMPIRequest_AS_REQUEST(o2)))
    res = Py_True;
  else
    res = Py_False;
  
  Py_INCREF(res);
  return res;
}


static int
request_nonzero(PyMPIRequestObject* ob) {
  return PyMPIRequest_AS_REQUEST(ob) != MPI_REQUEST_NULL;
}
static PyNumberMethods request_number_methods = {
	0,			/*nb_add*/
	0,			/*nb_subtract*/
	0,			/*nb_multiply*/
	0, 	 	 	/*nb_divide*/
	0,			/*nb_remainder*/
	0,			/*nb_divmod*/
	0,			/*nb_power*/
	0,			/*nb_negative*/
	0,			/*nb_positive*/
	0,			/*nb_absolute*/
	(inquiry)request_nonzero,	/*nb_nonzero*/
	0,			/*nb_invert*/
	0,			/*nb_lshift*/
	0,			/*nb_rshift*/
	0,			/*nb_and*/
	0,			/*nb_xor*/
	0,			/*nb_or*/
	0,			/*nb_coerce*/
	0,			/*nb_int*/
	0,			/*nb_long*/
	0,			/*nb_float*/
	0,			/*nb_oct*/
	0, 	 	 	/*nb_hex*/
	0,			/*nb_inplace_add*/
	0,			/*nb_inplace_subtract*/
	0,			/*nb_inplace_multiply*/
	0,			/*nb_inplace_divide*/
	0,			/*nb_inplace_remainder*/
	0,			/*nb_inplace_power*/
	0,			/*nb_inplace_lshift*/
	0,			/*nb_inplace_rshift*/
	0,			/*nb_inplace_and*/
	0,			/*nb_inplace_xor*/
	0,			/*nb_inplace_or*/
	0,			/*nb_floor_divide */
	0,			/*nb_true_divide */
	0,			/*nb_inplace_floor_divide */
	0,			/*nb_inplace_true_divide */
};
#define request_as_number   &request_number_methods
#define request_as_sequence (PySequenceMethods*)0
#define request_as_mapping  (PyMappingMethods*)0


#define request_methods 0
#define request_members 0
#define request_getset 0


PyDoc_STRVAR(request_doc, "Request type");

static PyTypeObject PyMPIRequest_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,					/*ob_size*/
  "libmpi.Request",			/*tp_name*/
  sizeof(PyMPIRequestObject),		/*tp_basicsize*/
  0,					/*tp_itemsize*/

  (destructor)request_dealloc,		/*tp_dealloc*/
  (printfunc)0,				/*tp_print*/
  (getattrfunc)0,			/*tp_getattr*/
  (setattrfunc)0,			/*tp_setattr*/
  (cmpfunc)0,				/*tp_compare*/
  (reprfunc)0,				/*tp_repr*/

  request_as_number,			/*tp_as_number*/
  request_as_sequence,			/*tp_as_sequence*/
  request_as_mapping,			/*tp_as_mapping*/

  (hashfunc)0,				/*tp_hash*/
  (ternaryfunc)0,			/*tp_call*/
  (reprfunc)0,				/*tp_str*/
  (getattrofunc)0,			/*tp_getattro*/
  (setattrofunc)0,			/*tp_setattro*/

  (PyBufferProcs*)0,			/*tp_as_buffer*/

  Py_TPFLAGS_DEFAULT  |
  Py_TPFLAGS_BASETYPE,			/*tp_flags*/

  request_doc, 				/*tp_doc*/

  (traverseproc)0,			/* tp_traverse */

  (inquiry)0,				/* tp_clear */

  (richcmpfunc)request_richcompare,	/* tp_richcompare */

  (long)0,				/* tp_weaklistoffset */

  (getiterfunc)0,			/* tp_iter */
  (iternextfunc)0,			/* tp_iternext */

  request_methods,			/* tp_methods */
  request_members,			/* tp_members */
  request_getset,				/* tp_getset */
  0,					/* tp_base */
  0,					/* tp_dict */
  (descrgetfunc)0,			/* tp_descr_get */
  (descrsetfunc)0,			/* tp_descr_set */
  (long)0,				/* tp_dictoffset */

  request_init,				/* tp_init */
  request_alloc,				/* tp_alloc */
  request_new,				/* tp_new */
  PyObject_Del,           		/* tp_free */
};



/* PyMPIRequest C API */

static int
PyMPIRequest_Check(PyObject *op)
{
  return PyObject_TypeCheck(op, &PyMPIRequest_Type);
}

static int
PyMPIRequest_CheckExact(PyObject *op)
{
  return ((op)->ob_type == &PyMPIRequest_Type);
}

static PyObject*
PyMPIRequest_FromRequest(MPI_Request request)
{
  PyMPIRequestObject *self;
  PyTypeObject *type = &PyMPIRequest_Type;
  self = (PyMPIRequestObject*) type->tp_alloc(type, 0);
  if (self != NULL) {
    self->request = (PyObject*) PyObject_New(_Py_MPI_Request,
					   &_Py_MPI_Request_Type);
    if (self->request == NULL) {
      Py_DECREF(self);
      return NULL;
    }
    PyMPIRequest_AS_REQUEST(self) = request;
  }
  return (PyObject*)self;
}

static MPI_Request
PyMPIRequest_AsRequest(PyObject *self)
{
  if (self) {
    if (PyMPIRequest_Check(self))
      return PyMPIRequest_AS_REQUEST(self);
    PyErr_SetString(PyExc_TypeError,
		    "argument must be a 'MPI_Request' object");
  }
  if (!PyErr_Occurred())
    PyErr_SetString(PyExc_TypeError,
		    "PyMPIRequest_AsRequest() called with null pointer");
  return MPI_REQUEST_DEFAULT;
}

static MPI_Request*
PyMPIRequest_AsRequestPtr(PyObject *self)
{
  if (self) {
    if (PyMPIRequest_Check(self))
      return PyMPIRequest_AS_REQUEST_PTR(self);
    PyErr_SetString(PyExc_TypeError,
		    "argument must be a 'MPI_Request' object");
  }
  if (!PyErr_Occurred())
    PyErr_SetString(PyExc_TypeError,
		    "PyMPIRequest_AsRequestPtr() called with null pointer");
  return NULL;
}


/* API pointers to export */

static PyTypeObject* PyMPIRequest_TYPE = &PyMPIRequest_Type;

static void* PyMPIRequest_API[] = {
  /*[0]*/ (void*) PyMPIRequest_Check,
  /*[1]*/ (void*) PyMPIRequest_CheckExact,
  /*[2]*/ (void*) PyMPIRequest_FromRequest,
  /*[3]*/ (void*) PyMPIRequest_AsRequest,
  /*[4]*/ (void*) PyMPIRequest_AsRequestPtr,
};

/* PyMPIStatus_Type implementation */

static PyTypeObject _Py_MPI_Status_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,				/*ob_size*/
  "_Py_MPI_Status",		/*tp_name*/
  sizeof(_Py_MPI_Status),	/*tp_basicsize*/
  0,				/*tp_itemsize*/
  (destructor)PyObject_Del,	/*tp_dealloc*/
  0,
};


static MPI_Status status_default(void)
{
  MPI_Status status;
  status.MPI_SOURCE = MPI_UNDEFINED;
  status.MPI_TAG    = MPI_ANY_TAG;
  status.MPI_ERROR  = MPI_SUCCESS;
  return status;
}
#define MPI_STATUS_DEFAULT status_default()

#define status_alloc PyType_GenericAlloc

static void
status_dealloc(PyMPIStatusObject *self)
{
  Py_XDECREF(self->status);
  self->ob_type->tp_free(self);
}

static PyObject *
status_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
  PyObject *ob = NULL;
  static char *kwlist[] = {"status", 0};
  
  if (!PyArg_ParseTupleAndKeywords(args, kwds, 
				   "|O:Status", kwlist,
				   &ob)) return NULL;
  
  if (ob == NULL || ob == Py_None) {
    PyMPIStatusObject *self;
    self = (PyMPIStatusObject*) type->tp_alloc(type, 0);
    if (self != NULL) {
      self->status = (PyObject*) PyObject_New(_Py_MPI_Status,
					     &_Py_MPI_Status_Type);
      if (self->status == NULL) {
	Py_DECREF(self);
	return NULL;
      }
      PyMPIStatus_AS_STATUS(self) = MPI_STATUS_DEFAULT;
    }
    return (PyObject*)self;
  }
  
  if (PyMPIStatus_Check(ob)) {
    PyMPIStatusObject *self;
    self = (PyMPIStatusObject*) type->tp_alloc(type, 0);
    if (self != NULL) {
      self->status = (PyObject*) PyObject_New(_Py_MPI_Status,
					      &_Py_MPI_Status_Type);
      if (self->status == NULL) {
	Py_DECREF(self);
	return NULL;
      }
      PyMPIStatus_AS_STATUS(self) = PyMPIStatus_AS_STATUS(ob);
    }
    return (PyObject*) self;
  }
  
  PyErr_SetString(PyExc_TypeError,
		  "Status() argument must be a 'MPI_Status' object");
  return NULL;
}

#define status_init (initproc)0

static PyObject *
status_richcompare(PyMPIStatusObject * o1, PyMPIStatusObject * o2, int op)
{
  PyObject *res;

  /* Make sure both arguments are the same type */
  if (!(PyMPIStatus_Check((PyObject*)o1) && 
	PyMPIStatus_Check((PyObject*)o2))) {
    Py_INCREF(Py_NotImplemented);
    return Py_NotImplemented;
  }
  PyErr_SetString(PyExc_TypeError,
		  "cannot compare 'MPI_Status' objects");
  res = NULL;
  return res;
}


#define status_as_number   (PyNumberMethods*)0
#define status_as_sequence (PySequenceMethods*)0
#define status_as_mapping  (PyMappingMethods*)0


#define status_methods 0
#define status_members 0
#define LIBMPI_STATUS_GETSETER(ATTRIB) \
static PyObject * \
status_get_##ATTRIB(PyMPIStatusObject *self, void *closure) \
{ \
  return PyInt_FromLong((long)(PyMPIStatus_AS_STATUS(self).MPI_##ATTRIB)); \
} \
static int \
status_set_##ATTRIB(PyMPIStatusObject *self, PyObject *value, void *closure) \
{ \
  if (value == NULL) { \
    PyErr_SetString(PyExc_TypeError, "cannot delete attribute"); \
    return -1; \
  } \
  if (!PyInt_Check(value)) { \
    PyErr_SetString(PyExc_TypeError, "attribute value must be a integer"); \
    return -1; \
  } \
  PyMPIStatus_AS_STATUS(self).MPI_##ATTRIB = (int) PyInt_AS_LONG(value); \
  return 0; \
}
LIBMPI_STATUS_GETSETER(SOURCE)
LIBMPI_STATUS_GETSETER(TAG)
LIBMPI_STATUS_GETSETER(ERROR)
static PyGetSetDef status_getset[] = {
  {"MPI_SOURCE",
   (getter)status_get_SOURCE, (setter)status_set_SOURCE,
   "message source", NULL},
  {"MPI_TAG",
   (getter)status_get_TAG,    (setter)status_set_TAG,
   "message tag",    NULL},
  {"MPI_ERROR",
   (getter)status_get_ERROR,  (setter)status_set_ERROR,
   "message error",  NULL},
  {NULL}  /* Sentinel */
};


PyDoc_STRVAR(status_doc, "Status type");

static PyTypeObject PyMPIStatus_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,					/*ob_size*/
  "libmpi.Status",			/*tp_name*/
  sizeof(PyMPIStatusObject),		/*tp_basicsize*/
  0,					/*tp_itemsize*/

  (destructor)status_dealloc,		/*tp_dealloc*/
  (printfunc)0,				/*tp_print*/
  (getattrfunc)0,			/*tp_getattr*/
  (setattrfunc)0,			/*tp_setattr*/
  (cmpfunc)0,				/*tp_compare*/
  (reprfunc)0,				/*tp_repr*/

  status_as_number,			/*tp_as_number*/
  status_as_sequence,			/*tp_as_sequence*/
  status_as_mapping,			/*tp_as_mapping*/

  (hashfunc)0,				/*tp_hash*/
  (ternaryfunc)0,			/*tp_call*/
  (reprfunc)0,				/*tp_str*/
  (getattrofunc)0,			/*tp_getattro*/
  (setattrofunc)0,			/*tp_setattro*/

  (PyBufferProcs*)0,			/*tp_as_buffer*/

  Py_TPFLAGS_DEFAULT  |
  Py_TPFLAGS_BASETYPE,			/*tp_flags*/

  status_doc, 				/*tp_doc*/

  (traverseproc)0,			/* tp_traverse */

  (inquiry)0,				/* tp_clear */

  (richcmpfunc)status_richcompare,	/* tp_richcompare */

  (long)0,				/* tp_weaklistoffset */

  (getiterfunc)0,			/* tp_iter */
  (iternextfunc)0,			/* tp_iternext */

  status_methods,			/* tp_methods */
  status_members,			/* tp_members */
  status_getset,				/* tp_getset */
  0,					/* tp_base */
  0,					/* tp_dict */
  (descrgetfunc)0,			/* tp_descr_get */
  (descrsetfunc)0,			/* tp_descr_set */
  (long)0,				/* tp_dictoffset */

  status_init,				/* tp_init */
  status_alloc,				/* tp_alloc */
  status_new,				/* tp_new */
  PyObject_Del,           		/* tp_free */
};



/* PyMPIStatus C API */

static int
PyMPIStatus_Check(PyObject *op)
{
  return PyObject_TypeCheck(op, &PyMPIStatus_Type);
}

static int
PyMPIStatus_CheckExact(PyObject *op)
{
  return ((op)->ob_type == &PyMPIStatus_Type);
}

static PyObject*
PyMPIStatus_FromStatus(MPI_Status status)
{
  PyMPIStatusObject *self;
  PyTypeObject *type = &PyMPIStatus_Type;
  self = (PyMPIStatusObject*) type->tp_alloc(type, 0);
  if (self != NULL) {
    self->status = (PyObject*) PyObject_New(_Py_MPI_Status,
					   &_Py_MPI_Status_Type);
    if (self->status == NULL) {
      Py_DECREF(self);
      return NULL;
    }
    PyMPIStatus_AS_STATUS(self) = status;
  }
  return (PyObject*)self;
}

static MPI_Status
PyMPIStatus_AsStatus(PyObject *self)
{
  if (self) {
    if (PyMPIStatus_Check(self))
      return PyMPIStatus_AS_STATUS(self);
    PyErr_SetString(PyExc_TypeError,
		    "argument must be a 'MPI_Status' object");
  }
  if (!PyErr_Occurred())
    PyErr_SetString(PyExc_TypeError,
		    "PyMPIStatus_AsStatus() called with null pointer");
  return MPI_STATUS_DEFAULT;
}

static MPI_Status*
PyMPIStatus_AsStatusPtr(PyObject *self)
{
  if (self) {
    if (PyMPIStatus_Check(self))
      return PyMPIStatus_AS_STATUS_PTR(self);
    PyErr_SetString(PyExc_TypeError,
		    "argument must be a 'MPI_Status' object");
  }
  if (!PyErr_Occurred())
    PyErr_SetString(PyExc_TypeError,
		    "PyMPIStatus_AsStatusPtr() called with null pointer");
  return NULL;
}


/* API pointers to export */

static PyTypeObject* PyMPIStatus_TYPE = &PyMPIStatus_Type;

static void* PyMPIStatus_API[] = {
  /*[0]*/ (void*) PyMPIStatus_Check,
  /*[1]*/ (void*) PyMPIStatus_CheckExact,
  /*[2]*/ (void*) PyMPIStatus_FromStatus,
  /*[3]*/ (void*) PyMPIStatus_AsStatus,
  /*[4]*/ (void*) PyMPIStatus_AsStatusPtr,
};


/*------------------------------------------------------------------*/

/* Helper macro to initialize MPI types */
#define LIBMPI_TYPE_INIT(Type)                            \
{ if (PyType_Ready(&_Py_MPI_##Type##_Type) < 0) return; } \
{ if (PyType_Ready(&PyMPI##Type##_Type)    < 0) return; }

/* Helper macro to add MPI types and API pointers */
#define LIBMPI_TYPE_ADD(m, Type)                                         \
{                                                                        \
  PyObject* tp  = (PyObject*) PyMPI##Type##_TYPE;                        \
  PyObject* api = PyCObject_FromVoidPtr((void*)PyMPI##Type##_API, NULL); \
  if (api == NULL) return;                                               \
  if (PyModule_AddObject(m, #Type"Type", tp)         < 0) return;        \
  if (PyModule_AddObject(m, "PyMPI"#Type"_API", api) < 0) return;        \
}

/*------------------------------------------------------------------*/

PyDoc_STRVAR(libmpi_doc, "MPI Types and C API");

static struct PyMethodDef libmpi_methods[] = {
  {NULL, NULL}        /* Sentinel */
};

PyMODINIT_FUNC
initlibmpi(void) {

  PyObject* m;

  LIBMPI_TYPE_INIT( Op         );
  LIBMPI_TYPE_INIT( Comm       );
  LIBMPI_TYPE_INIT( Group      );
  LIBMPI_TYPE_INIT( Status     );
  LIBMPI_TYPE_INIT( Request    );
  LIBMPI_TYPE_INIT( Datatype   );
  LIBMPI_TYPE_INIT( Errhandler );
  
  m = Py_InitModule4("mpi4py.libmpi",
		     libmpi_methods, libmpi_doc,
		     NULL, PYTHON_API_VERSION);
  if (m == NULL) return;

  LIBMPI_TYPE_ADD(m, Op         );
  LIBMPI_TYPE_ADD(m, Comm       );
  LIBMPI_TYPE_ADD(m, Group      );
  LIBMPI_TYPE_ADD(m, Status     );
  LIBMPI_TYPE_ADD(m, Request    );
  LIBMPI_TYPE_ADD(m, Datatype   );
  LIBMPI_TYPE_ADD(m, Errhandler );

}

/*------------------------------------------------------------------*/

/* 
 * Local Variables:
 * mode: C
 * End:
 */
