/* $Id: macros.h,v 1.4 2006/08/10 22:43:48 dalcinl Exp $ */

#ifndef Py_MPI4PY_MACROS_H
#define Py_MPI4PY_MACROS_H

#include <Python.h>

#define TRY
#define RAISE  goto fail
#define EXCEPT fail:

#define PY_SET_ERROR(pyexc,msg) \
{ if((pyexc) != NULL) PyErr_SetString((pyexc), (msg)?(msg):""); }

#define MPI_SET_ERROR(ierr) \
{ PyErr_SetObject(PyExc_MPIError, PyInt_FromLong((long)(ierr))); }

#define PY_RAISE(pyexc,msg) { PY_SET_ERROR(pyexc,msg); RAISE; }

#define MPI_RAISE(ierr)     { MPI_SET_ERROR(ierr); RAISE; }


#define MPI_CHECK_NULL(object,nullhandle,errclass) \
if ( (object) == (nullhandle) ) { \
  MPI_SET_ERROR(errclass); \
  return NULL; \
}


#define ERRHANDLER_FREE(errhdl) { \
    if (errhdl != MPI_ERRORS_ARE_FATAL && \
	errhdl != MPI_ERRORS_RETURN) \
      MPI_Errhandler_free(&errhdl); \
    else { \
      errhdl = MPI_ERRHANDLER_NULL; \
    } \
}

/* Hack for MPICH !!! */

#if defined(MPICH_NAME)
#undef  ERRHANDLER_FREE
#define ERRHANDLER_FREE(errhdl) { \
   MPI_Errhandler_free(&errhdl);  \
}
#endif

/* Hack for LAM !!! */

#if defined(LAM_MPI)
struct _errhdl {
  void  (*eh_func)(void);
  int   eh_refcount;
  int   eh_f77handle;
  int   eh_flags;
};
#undef  ERRHANDLER_FREE
#define ERRHANDLER_FREE(errhdl) { \
    if (errhdl != MPI_ERRORS_ARE_FATAL && \
	errhdl != MPI_ERRORS_RETURN) \
      MPI_Errhandler_free(&errhdl); \
    else { \
      errhdl->eh_refcount--; \
      errhdl = MPI_ERRHANDLER_NULL; \
    } \
}
#endif

/* Hack for Open MPI !!! */

#if defined(OPEN_MPI)
#undef  ERRHANDLER_FREE
#define ERRHANDLER_FREE(errhdl) { \
    if (errhdl != MPI_ERRORS_ARE_FATAL && \
	errhdl != MPI_ERRORS_RETURN) \
      MPI_Errhandler_free(&errhdl); \
    else { \
      errhdl = MPI_ERRHANDLER_NULL; \
    } \
}
#endif


#define MPI_CHECK(ierr) { \
  int _ierr_ = (ierr); \
  if (_ierr_ != MPI_SUCCESS) MPI_RAISE(_ierr_); \
}


#define MPI_CHECK_CALL(mpi_call) { \
  int _ierr_; \
  MPI_Errhandler _errhdl_; \
  MPI_Errhandler_get(MPI_COMM_WORLD, &_errhdl_); \
  MPI_Errhandler_set(MPI_COMM_WORLD, MPI_ERRORS_RETURN); \
  Py_BEGIN_ALLOW_THREADS \
  _ierr_ = (mpi_call); \
  Py_END_ALLOW_THREADS \
  MPI_Errhandler_set(MPI_COMM_WORLD, _errhdl_); \
  ERRHANDLER_FREE(_errhdl_); \
  if (_ierr_ != MPI_SUCCESS) MPI_RAISE(_ierr_); \
}

#define MPI_CHECK_CALL_COMM(comm,comm_call) { \
  int _ierr_; \
  MPI_Errhandler _errhdl_; \
  if (comm == MPI_COMM_NULL) MPI_RAISE(MPI_ERR_COMM); \
  MPI_Errhandler_get(comm, &_errhdl_); \
  MPI_Errhandler_set(comm, MPI_ERRORS_RETURN); \
  Py_BEGIN_ALLOW_THREADS \
  _ierr_ = (comm_call); \
  Py_END_ALLOW_THREADS \
  MPI_Errhandler_set(comm, _errhdl_); \
  ERRHANDLER_FREE(_errhdl_); \
  if (_ierr_ != MPI_SUCCESS) MPI_RAISE(_ierr_); \
}

#define MPI_MAKE_NEWCOMM(comm,newcomm,call) { \
  int _ierr_; \
  MPI_Errhandler _errhdl_; \
  if (comm == MPI_COMM_NULL) MPI_RAISE(MPI_ERR_COMM); \
  MPI_Errhandler_get(comm, &_errhdl_); \
  MPI_Errhandler_set(comm, MPI_ERRORS_RETURN); \
  Py_BEGIN_ALLOW_THREADS \
  _ierr_ = (call); \
  Py_END_ALLOW_THREADS \
  MPI_Errhandler_set(comm, _errhdl_); \
  if (_ierr_ != MPI_SUCCESS) \
    { ERRHANDLER_FREE(_errhdl_); MPI_RAISE(_ierr_); } \
  if (newcomm != MPI_COMM_NULL) \
    { MPI_Errhandler_set(newcomm, _errhdl_); } \
  ERRHANDLER_FREE(_errhdl_); \
}

#define MPI_CHECK_CALL_FREE(mpi_call) { \
  int _ierr_; \
  MPI_Errhandler _errhdl_; \
  MPI_Errhandler_get(MPI_COMM_WORLD, &_errhdl_); \
  MPI_Errhandler_set(MPI_COMM_WORLD, MPI_ERRORS_RETURN); \
  _ierr_ = (mpi_call); \
  MPI_Errhandler_set(MPI_COMM_WORLD, _errhdl_); \
  ERRHANDLER_FREE(_errhdl_); \
  if (_ierr_ != MPI_SUCCESS) MPI_RAISE(_ierr_); \
}


#if !defined(Py_RETURN_NONE)
#define Py_RETURN_NONE return Py_INCREF(Py_None), Py_None
#endif

#if !defined(Py_BOOLOBJECT_H)
#define PyBool_FromLong(b) PyInt_FromLong(((long)(b))?1:0)
#endif


#endif /* !Py_MPI4PY_MACROS_H */
