import unittest
from mpi4py import MPI

class TestEnviron(unittest.TestCase):

    def testIsInitFini(self):
        self.assertTrue(MPI.Is_initialized())
        self.assertFalse(MPI.Is_finalized())

    def testVersion(self):
        version = MPI.Get_version()
        self.assertEqual(len(version), 2)
        self.assertTrue(type(version[0]) is int)
        self.assertTrue(type(version[1]) is int)
        
    def testTime(self):
        tick = MPI.Wtick()
        time1 = MPI.Wtime()
        time2 = MPI.Wtime()
        self.assertTrue(time1 < time2)
        self.assertTrue((time2 - time1) >= tick)

    def testProcessorName(self):
        procname = MPI.Get_processor_name()
        self.assertTrue(type(procname[1]) is str)

    def testHostPorcessor(self):
        size = MPI.COMM_WORLD.Get_size()
        vals = range(size) + [MPI.UNDEFINED]
        self.assertTrue(MPI.HOST in vals)

    def testIOProcessor(self):
        size = MPI.COMM_WORLD.Get_size()
        vals = range(size) + [MPI.UNDEFINED]
        self.assertTrue(MPI.IO in vals)
        
if __name__ == '__main__':
    unittest.main()
